/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_INPUTCOMPONENTS_HPP
#define MYFITTER_INPUTCOMPONENTS_HPP

#include "inputfunction.hpp"


namespace myfitter {

// single observables with gaussian and systematic errors.

class GaussianIC : public InputComponent {
private:
    double stat_err_;
    double syst_minus_;
    double syst_plus_;

public:
    // constructors
    GaussianIC(int iobs, double mean, double stat_err,
               double syst_plus=0., double syst_minus=0.) :
        InputComponent(iobs+1), stat_err_(stat_err),
        syst_minus_(syst_minus), syst_plus_(syst_plus) {
        claim_(iobs, mean);
    }
    GaussianIC(const GaussianIC& lc) :
        InputComponent(lc), stat_err_(lc.stat_err_),
        syst_minus_(lc.syst_minus_), syst_plus_(lc.syst_plus_) { }
    virtual GaussianIC* copy() const
    { return new GaussianIC(*this); }
    
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result);

    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs) {
        hess(0, 0) = 1./(stat_err_*stat_err_);
    }
};


// asymmetric Gaussian

class AsymmetricGaussianIC : public InputComponent {
private:
    double stat_minus_;
    double stat_plus_;
    double syst_minus_;
    double syst_plus_;

public:
    // constructors
    AsymmetricGaussianIC(int iobs, double mean,
                         double stat_plus, double stat_minus,
                         double syst_plus=0., double syst_minus=0.) :
        InputComponent(iobs+1),
        stat_minus_(stat_minus), stat_plus_(stat_plus),
        syst_minus_(syst_minus), syst_plus_(syst_plus)
    { claim_(iobs, mean); }
    AsymmetricGaussianIC(const AsymmetricGaussianIC& lc) :
        InputComponent(lc),
        stat_minus_(lc.stat_minus_), stat_plus_(lc.stat_plus_),
        syst_minus_(lc.syst_minus_), syst_plus_(lc.syst_plus_) { }
    virtual AsymmetricGaussianIC* copy() const
    { return new AsymmetricGaussianIC(*this); }
    
    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result);

    virtual double calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs) {
        hess(0, 0) = 1./std::pow(0.5*(stat_plus_ + stat_minus_), 2);
    }
};


// correlated observables with gaussian errors

class CorrelatedGaussianIC : public InputComponent {
private:
    ObservableVector err_;
    boost::numeric::ublas::matrix<double> cor_, invvar_;
    bool needs_update_;

    void update_();

public:
    CorrelatedGaussianIC(int nobs) 
        : InputComponent(nobs), needs_update_(true) { }
    CorrelatedGaussianIC(const CorrelatedGaussianIC& c) :
        InputComponent(c), err_(c.err_), cor_(c.cor_),
        invvar_(c.invvar_), needs_update_(c.needs_update_) { }
    virtual CorrelatedGaussianIC* copy() const
    { return new CorrelatedGaussianIC(*this); }

    // claim observable with index iobs
    void add(int iobs, double val, double err);
    void add(int iobs, double val) { add(iobs, val, Model::nan()); }
    void add(int iobs) { add(iobs, Model::nan(), Model::nan()); }
    
    // get/set central value of observable with external index iobs
    double central_value(int iobs) const
    { return central_value_(internal_index(iobs)); }
    void central_value(int iobs, double x) 
    { central_value_(internal_index(iobs), x); }

    // get/set standard deviation of observable with external index iobs
    double sigma(int iobs) const
    { return err_[internal_index(iobs)]; }
    void sigma(int iobs, double x)
    { err_[internal_index(iobs)] = x; needs_update_ = true; }

    // get/set correlation of observables i and j
    double cor(int i, int j) const {
        return cor_(internal_index(i), internal_index(j));
    }
    void cor(int i, int j, double x) {
        if(i != j) {
            cor_(internal_index(i), internal_index(j))
                = cor_(internal_index(j), internal_index(i)) = x;
            needs_update_ = true;
        }
    }

    virtual double calc(const ObservableVector& pred,
                        const ObservableVector& meas);

    virtual int calc_deriv(const ObservableVector& pred,
                              const ObservableVector& meas,
                              ObservableVector& result);

    virtual void get_hessian(Matrix& hess, const ObservableVector& obs);
};

}

#endif
