/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_MODEL_HPP
#define MYFITTER_MODEL_HPP

#include <vector>
#include <algorithm>
#include <limits>
#include <cassert>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <dvegas/dvegas.h>


namespace myfitter {

typedef boost::numeric::ublas::matrix<double> Matrix;
typedef boost::numeric::ublas::vector<double> Vector;
typedef Vector ParameterVector;
typedef Vector ObservableVector;


// base class for models

class Model {
private:
    int npar_;
    int nobs_;
    ParameterVector par_min_, par_max_;
    ParameterVector par_val_, par_scale_;
    std::vector<bool> fixed_, smallrange_;
    ParameterVector scan_min_, scan_max_;
    std::vector<ParameterVector> par_map_;
    std::vector<ObservableVector> obs_map_;
    double chisquare_;
    double constraint_penalty_;
    double derivative_epsilon_;
    unsigned int random_seed_;
    boost::mt19937 base_rng_;

protected:
    ObservableVector observables_;
    Matrix derivatives_;
    bool needs_update_;

    int numerical_derivative_(int i);

public:
    // constructor
    // npar: number of parameters of the model
    // nobs: number of observables
    Model(int npar, int nobs) :
        npar_(npar), nobs_(nobs), par_min_(npar, Model::nan()),
        par_max_(npar, Model::nan()), par_val_(npar, Model::nan()),
        par_scale_(npar, Model::nan()), fixed_(npar, false),
        smallrange_(npar, false), scan_min_(npar, Model::nan()),
        scan_max_(npar, Model::nan()),
        chisquare_(Model::nan()), constraint_penalty_(Model::nan()),
        derivative_epsilon_(1.e-5),
        random_seed_(1234u), base_rng_(random_seed_),
        observables_(nobs, Model::nan()),
        derivatives_(npar, nobs, Model::nan()), needs_update_(true) {  }

    // copy constructor
    Model(const Model& m) :
        npar_(m.npar_), nobs_(m.nobs_),
        par_min_(m.par_min_), par_max_(m.par_max_), par_val_(m.par_val_),
        par_scale_(m.par_scale_), fixed_(m.fixed_), smallrange_(m.smallrange_), 
        scan_min_(m.scan_min_), scan_max_(m.scan_max_),
        par_map_(m.par_map_), obs_map_(m.obs_map_), chisquare_(m.chisquare_),
        constraint_penalty_(m.constraint_penalty_),
        derivative_epsilon_(m.derivative_epsilon_),
        random_seed_(m.random_seed_), base_rng_(m.base_rng_),
        observables_(m.observables_), derivatives_(m.derivatives_),
        needs_update_(m.needs_update_) { }

    virtual Model& operator=(const Model& m) {
        npar_ = m.npar_;
        nobs_ = m.nobs_;
        par_min_ = m.par_min_;
        par_max_ = m.par_max_;
        par_val_ = m.par_val_;
        par_scale_ = m.par_scale_;
        fixed_ = m.fixed_;
        smallrange_ = m.smallrange_;
        scan_min_ = m.scan_min_;
        scan_max_ = m.scan_max_;
        par_map_ = m.par_map_;
        obs_map_ = m.obs_map_;
        chisquare_ = m.chisquare_;
        constraint_penalty_ = m.constraint_penalty_;
        derivative_epsilon_ = m.derivative_epsilon_;
        random_seed_ = m.random_seed_;
        base_rng_ = m.base_rng_;
        observables_ = m.observables_;
        derivatives_ = m.derivatives_;
        needs_update_ = m.needs_update_;
        return *this;
    }

    // number of parameters
    int nparameters() const { return npar_; }

    // number of observables
    int nobservables() const { return nobs_; }

    // returns nan. Can be used to 'unset' parameter limits.
    // A limit of nan means the parameter is unbounded in that direction.
    static double nan() { return std::numeric_limits<double>::quiet_NaN(); }

    // get/set the small parameter used to calculate derivatives numerically.
    // The derivative with respect to parameter i is calculated numerically by
    // shifting the parameter by derivative_epsilon()*scale(i).
    double derivative_epsilon() const { return derivative_epsilon_; }
    void derivative_epsilon(double e) { derivative_epsilon_ = e; }

    // get/set the value of parameter i.
    // Values of fixed parameters can not be changed.
    double parameter(int i) const
    { assert(i >= 0 && i < nparameters()); return par_val_[i]; }
    void parameter(int i, double x) {
        assert(i >= 0 && i < nparameters());
        if(fixed_[i]) return;
        if(!std::isnan(par_min_[i]) && x < par_min_[i]) 
            { parameter(i, par_min_[i]); return; }
        if(!std::isnan(par_max_[i]) && x > par_max_[i])
            { parameter(i, par_max_[i]); return; }
        if(par_val_[i] != x) {
            par_val_[i] = x;
            needs_update_ = true;
        }
    }

    // set parameter value and scale at once
    void parameter(int i, double x, double dx)
    { parameter(i, x); scale(i, dx); }

    // set parameter value, scale and limits at once
    void parameter(int i, double x, double dx, double xmin, double xmax)
    { parameter(i, x); scale(i, dx); range(i, xmin, xmax); }

    // get/set parameter vector
    const ParameterVector& parameters() { return par_val_; }
    void parameters(const ParameterVector& par) {
        assert(int(par.size()) == nparameters());
        for(int i=0; i<nparameters(); ++i) parameter(i, par[i]);
    }

    // fix/release parameter i.
    void fix(int i)
    { assert(i >= 0 && i < nparameters()); fixed_[i] = true; }
    void release(int i)
    { assert(i >= 0 && i < nparameters()); fixed_[i] = false; }
    bool fixed(int i) const
    { assert(i >= 0 && i < nparameters()); return fixed_[i]; }

    // get/set smallrange flag for parameter i
    // In a p-value integration the smallrange flag indicates that the
    // parameter should be considered fixed when determining the model's
    // hyperplane in sample space
    bool smallrange(int i)
    { assert(i >= 0 && i < nparameters()); return smallrange_[i]; }
    void smallrange(int i, bool b)
    { assert(i >= 0 && i < nparameters()); smallrange_[i] = b; }

    // get/set typical scale of parameter i.
    double scale(int i) const
    { assert(i >= 0 && i < nparameters()); return par_scale_[i]; }
    void scale(int i, double x)
    { assert(i >= 0 && i < nparameters()); par_scale_[i] = x; }

    // get/set upper and lower limits on parameter i. In a fit, the parameter
    // is then only allowed to float within these limits.
    // A limit of nan means the parameter is unbounded in that direction.
    double upper_limit(int i) const
    { assert(i >= 0 && i < nparameters()); return par_max_[i]; }
    void upper_limit(int i, double x) {
        assert(i >= 0 && i < nparameters()); par_max_[i] = x;
        if(par_val_[i] > x) par_val_[i] = x;
    }
    double lower_limit(int i) const
    { assert(i >= 0 && i < nparameters()); return par_min_[i]; }
    void lower_limit(int i, double x) {
        assert(i >= 0 && i < nparameters()); par_min_[i] = x;
        if(par_val_[i] < x) par_val_[i] = x;
    }
    void range(int i, double x, double y) {
        assert(i >= 0 && i < nparameters()); par_min_[i] = x; par_max_[i] = y;
        if(par_val_[i] < x) par_val_[i] = x;
        if(par_val_[i] > y) par_val_[i] = y;
    }

    // unset upper or lower limits
    void unset_upper_limit(int i)
    { assert(i >= 0 && i < nparameters()); par_max_[i] = nan(); }
    void unset_lower_limit(int i)
    { assert(i >= 0 && i < nparameters()); par_min_[i] = nan(); }
    void unset_limits(int i)
    { assert(i >= 0 && i < nparameters()); par_max_[i] = par_min_[i] = nan(); }

    // check if parameter i has upper/lower limit
    bool has_upper_limit(int i)
    { assert(i >= 0 && i < nparameters()); return !std::isnan(par_max_[i]); }
    bool has_lower_limit(int i)
    { assert(i >= 0 && i < nparameters()); return !std::isnan(par_min_[i]); }

    // get/set the scan range for parameter i. These ranges are used by the
    // scan() method.
    double scan_min(int i) const
    { assert(i >= 0 && i < nparameters()); return scan_min_[i]; }
    void scan_min(int i, double x) {
        assert(i >= 0 && i < nparameters());
        scan_min_[i] = has_lower_limit(i) ? std::max(x, lower_limit(i)) : x;
    }
    double scan_max(int i) const
    { assert(i >= 0 && i < nparameters()); return scan_max_[i]; }
    void scan_max(int i, double x) {
        assert(i >= 0 && i < nparameters());
        scan_max_[i] = has_upper_limit(i) ? std::min(x, upper_limit(i)) : x;
    }
    void scan_range(int i, double x, double y) {
        assert(i >= 0 && i < nparameters());
        scan_min_[i] = has_lower_limit(i) ? std::max(x, lower_limit(i)) : x;
        scan_max_[i] = has_upper_limit(i) ? std::min(y, upper_limit(i)) : y;
    }

    // set all scan ranges to current parameter value plus/minus factor times
    // the scale of that parameter
    void set_scan_ranges(double factor) {
        for(int i=0; i<nparameters(); ++i)
            scan_range(i, parameter(i)-factor*scale(i),
                       parameter(i)+factor*scale(i));
    }

    // get/set random seed for scans
    unsigned int random_seed() const { return random_seed_; }
    void random_seed(unsigned int rs) { random_seed_ = rs; }

    // clear dictionary of sample points and reset random number generator
    void clear()
    { base_rng_.seed(random_seed_); par_map_.clear();  obs_map_.clear(); }

    // sample the parameter space with n points with a uniform random
    // distribution over the specified ranges (scan_min, scan_max etc.).
    // The collected data is used by global_fit() to find the best start value.
    // Returns zero on success and the return value of the first failed call
    // to calc() on failure.
    int scan(int n);

    // store the current parameters and observables in the dictionary.
    // Observables are computed if an update is necessary. Returns 0 on
    // success and the return value of the calc() method on failure.
    int store();

    // returns the number of sample points from the last call to scan()
    int nsamples() const { return par_map_.size(); }    
    // returns the i-th sample parameter vector
    const ParameterVector& sample_parameters(int i) const
    { assert(i < int(par_map_.size())); return par_map_[i]; }
    // returns the i-th sample observable vector
    const ObservableVector& sample_observables(int i) const
    { assert(i < int(obs_map_.size())); return obs_map_[i]; }

    // get/set the chi-square value of the last fit with this model
    double chisquare() const { return chisquare_; }
    void chisquare(double csq) { chisquare_ = csq; }

    // get/set the constraint penalty value of the last fit with this model
    double constraint_penalty() const { return constraint_penalty_; }
    void constraint_penalty(double d) { constraint_penalty_ = d; }

    // read observable vector
    const ObservableVector& observables() const { return observables_; }

    // read individual observable
    double observable(int i) const
    { assert(i < nobservables()); assert(i >= 0); return observables_[i]; }

    // read derivative matrix
    // first index: parameter, second index: observable
    const Matrix& derivatives() const { return derivatives_; }

    // returns true if the derivatives calculated by calc_deriv() agree
    // with numerically calculated derivatives within a relative precision
    // rprec and an absolute precision aprec. On failure, it prints information
    // about the disagreeing derivatives to stderr.
    bool check_derivatives(double rprec, double aprec);

    // calculate observables from current parameter values
    // DERIVED CLASSES MUST OVERLOAD THIS!!
    virtual int calc() { needs_update_ = false; return 0; }

    // calculate derivatives of observables with respect to parameters
    virtual int calc_deriv() {
        int status;
        for(int i=0; i<nparameters(); ++i)
            if((status=numerical_derivative_(i))) return status;
        return 0;
    }
};


}

#endif
