/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012, 2013  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_UBLASUTILS_EIGEN_HPP
#define MYFITTER_UBLASUTILS_EIGEN_HPP

#include <boost/numeric/ublas/matrix_expression.hpp>
#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/exception.hpp>
#include <iostream>

namespace myfitter { namespace ublasutils { 

    template<class realT>
    inline realT modsq(std::complex<realT> z) {
        return real(z)*real(z)+imag(z)*imag(z);
    }


    template<class realT, class L, class A>
    void hh_tridiag_herm(boost::numeric::ublas::
                         matrix<std::complex<realT>, L, A>& M,
                         boost::numeric::ublas::
                         matrix<std::complex<realT>, L, A>& O,
                         realT prec = 1e-20)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::identity_matrix;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;
        using boost::numeric::ublas::prod;

        typedef std::complex<realT> complT;

        typedef matrix<complT, L, A> compl_mat;
        typedef vector<complT, A> compl_vec;

        typedef typename compl_mat::size_type size_type;

        typedef matrix_range<compl_mat> compl_mat_range;
        typedef matrix_column<compl_mat> compl_mat_col;
        typedef matrix_row<compl_mat> compl_mat_row;
        typedef vector_range<compl_mat_col> compl_mat_col_range;
        typedef vector_range<compl_mat_row> compl_mat_row_range;

        size_type N = M.size1();    
        BOOST_UBLAS_SAME(M.size1(), O.size1());
        BOOST_UBLAS_SAME(N, O.size2());
        BOOST_UBLAS_SAME(N, M.size2());

        compl_mat P(N,N);

        for(size_type n = N-1; n > 1; n--) {
            compl_mat_range Mn(M, range(0,n), range(0,n));
            compl_mat_range Pn(P, range(0,n), range(0,n));
            compl_mat_range On(O, range(0,N), range(0,n));

            compl_mat_col Mcol(M, n);
            compl_mat_row Mrow(M, n);
            compl_mat_col_range Mcoln(Mcol, range(0, n));
            compl_mat_row_range Mrown(Mrow, range(0, n));

            compl_vec u(n);
            u.assign(Mcoln);

            realT modunm1 = std::abs(u(n-1));
            if(modunm1 < prec)           
                u(n-1) += norm_2(u);
            else
                u(n-1) += (norm_2(u)/modunm1)*u(n-1);

            realT modu = norm_2(u);
            if(modu < prec) continue;
            u /= modu;
        
            Pn.assign(identity_matrix<complT>(n));
            Pn -= 2.*outer_prod(u, herm(u));
            On = prod(On, Pn);

            Mcoln = prod(Pn, Mcoln);
            Mrown = herm(Mcoln);

            compl_vec p(n), q(n);
            p = prod(Mn, u);

            q = 2.*p - (2.*inner_prod(herm(u), p))*u;
            Mn -= outer_prod(q, herm(u)) + outer_prod(u, herm(q));
        }
    }


    template<class realT, class L, class A>
    realT jacobi_diag_herm(boost::numeric::ublas::
                           matrix<std::complex<realT>, L, A>& M,
                           boost::numeric::ublas::
                           matrix<std::complex<realT>, L, A>& U,
                           unsigned int maxiter = 256,
                           realT prec = 1e-15)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::identity_matrix;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;
        using boost::numeric::ublas::prod;

        typedef std::complex<realT> complT;

        typedef matrix<complT, L, A> compl_mat;
        typedef vector<complT, A> compl_vec;

        typedef typename compl_mat::size_type size_type;

        typedef matrix_column<compl_mat> compl_mat_col;
        typedef matrix_row<compl_mat> compl_mat_row;

        size_type N = M.size1();    
        BOOST_UBLAS_SAME(M.size1(), U.size1());
        BOOST_UBLAS_SAME(N, U.size2());
        BOOST_UBLAS_SAME(N, M.size2());

        realT s=0, sdiag=0;
        size_type iter;
        size_type p,q;
        realT prec2 = 2.*prec/(N*(N-1));

        for(iter = 0; iter < maxiter; iter++) {
            // sum of off-diagonal squared elements
            s = 0.;
            sdiag = 0.;
            for(p = 0; p < N; p++) {
                sdiag += modsq(M(p,p));
                for(q = p+1; q < N; q++)
                    s += modsq(M(p,q));
            }
            if(s < prec*sdiag) break;


            for(p = 0; p < N; p++) {
                for(q = p+1; q < N; q++) {

                    // eigenvalues of the p,q submatrix

                    realT mpp = real(M(p,p));
                    realT mqq = real(M(q,q));

                    complT mpq = M(p,q);
                    realT modsqmpq = modsq(mpq);
                    if(modsqmpq < prec2*sdiag) continue;

                    realT mplus = (mpp + mqq)/2.;
                    realT mminus = (mpp - mqq)/2.;
                    realT root = std::sqrt(mminus*mminus + modsqmpq);

                    realT lamp = mplus - root;
                    realT lamq = mplus + root;


                    // eigenvectors

                    complT upp, upq, uqp, uqq, z, modv;

                    realT lampmpp = lamp - mpp;
                    realT lamqmpp = lamq - mpp;
                    realT lampmqq = lamp - mqq;
                    realT lamqmqq = lamq - mqq;

                    if(std::abs(lampmpp) > std::abs(lampmqq)) {
                        z = mpq/lampmpp;
                        modv = std::sqrt(modsq(z)+1.);
                        upp = z/modv;
                        uqp = 1./modv;
                    } else {
                        z = conj(mpq)/lampmqq;
                        modv = std::sqrt(modsq(z)+1.);
                        upp = 1./modv;
                        uqp = z/modv;
                    }
                    if(std::abs(lamqmpp) > std::abs(lamqmqq)) {
                        z = mpq/lamqmpp;
                        modv = std::sqrt(modsq(z)+1.);
                        upq = z/modv;
                        uqq = 1./modv;
                    } else {
                        z = conj(mpq)/lamqmqq;
                        modv = std::sqrt(modsq(z)+1.);
                        upq = 1./modv;
                        uqq = z/modv;
                    }


                    // matrix multiplications
                    
                    compl_mat_col Mcolp(M, p);
                    compl_mat_col Mcolq(M, q);
                    compl_mat_row Mrowp(M, p);
                    compl_mat_row Mrowq(M, q);
                    compl_mat_col Ucolp(U, p);
                    compl_mat_col Ucolq(U, q);

                    compl_vec xp(N), xq(N);
            
                    xp.assign(upp*Mcolp + uqp*Mcolq);
                    xq.assign(upq*Mcolp + uqq*Mcolq);
                    Mcolp.assign(xp);
                    Mcolq.assign(xq);

                    xp.assign(conj(upp)*Mrowp + conj(uqp)*Mrowq);
                    xq.assign(conj(upq)*Mrowp + conj(uqq)*Mrowq);
                    Mrowp.assign(xp);
                    Mrowq.assign(xq);

                    xp.assign(upp*Ucolp + uqp*Ucolq);
                    xq.assign(upq*Ucolp + uqq*Ucolq);
                    Ucolp.assign(xp);
                    Ucolq.assign(xq);
                }
            }
        }

        return s/sdiag;
    }


    // Jacobi diagonalization: Udag M U = diagonal

    template<class realT, class L, class A>
    realT jacobi_diag_symm(boost::numeric::ublas::matrix<realT, L, A>& M,
                           boost::numeric::ublas::matrix<realT, L, A>& U,
                           unsigned int maxiter = 256,
                           realT prec = 1e-15)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::identity_matrix;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;
        using boost::numeric::ublas::prod;

        typedef matrix<realT, L, A> real_mat;
        typedef vector<realT, A> real_vec;

        typedef typename real_mat::size_type size_type;

        typedef matrix_column<real_mat> real_mat_col;
        typedef matrix_row<real_mat> real_mat_row;

        size_type N = M.size1();    
        BOOST_ASSERT(M.size1() == U.size1());
        BOOST_ASSERT(N == U.size2());
        BOOST_ASSERT(N == M.size2());

        realT s=0, sdiag=0;
        size_type iter;
        size_type p,q;
        realT prec2 = 2.*prec/(N*(N-1));

        for(iter = 0; iter < maxiter; iter++) {
            // sum of off-diagonal squared elements
            s = 0;
            sdiag = 0.;
            for(p = 0; p < N; p++) {
                sdiag += M(p,p)*M(p,p);
                for(q = p+1; q < N; q++)
                    s += M(p,q)*M(p,q);
            }
            if(s < prec*sdiag) break;


            for(p = 0; p < N; p++) {
                for(q = p+1; q < N; q++) {

                    // eigenvalues of the p,q submatrix

                    realT mpp = M(p,p);
                    realT mqq = M(q,q);

                    realT mpq = M(p,q);
                    realT modsqmpq = mpq*mpq;
                    if(modsqmpq < prec2*sdiag) continue;

                    realT mplus = (mpp + mqq)/2.;
                    realT mminus = (mpp - mqq)/2.;
                    realT root = std::sqrt(mminus*mminus + modsqmpq);

                    realT lamp = mplus - root;
                    realT lamq = mplus + root;


                    // eigenvectors

                    realT upp, upq, uqp, uqq, z, modv;

                    realT lampmpp = lamp - mpp;
                    realT lamqmpp = lamq - mpp;
                    realT lampmqq = lamp - mqq;
                    realT lamqmqq = lamq - mqq;

                    if(lampmpp > lampmqq) {
                        z = mpq/lampmpp;
                        modv = std::sqrt(z*z+1.);
                        upp = z/modv;
                        uqp = 1./modv;
                    } else {
                        z = mpq/lampmqq;
                        modv = std::sqrt(z*z+1.);
                        upp = 1./modv;
                        uqp = z/modv;
                    }
                    if(lamqmpp > lamqmqq) {
                        z = mpq/lamqmpp;
                        modv = std::sqrt(z*z+1.);
                        upq = z/modv;
                        uqq = 1./modv;
                    } else {
                        z = mpq/lamqmqq;
                        modv = std::sqrt(z*z+1.);
                        upq = 1./modv;
                        uqq = z/modv;
                    }


                    // matrix multiplications
                    
                    real_mat_col Mcolp(M, p);
                    real_mat_col Mcolq(M, q);
                    real_mat_row Mrowp(M, p);
                    real_mat_row Mrowq(M, q);
                    real_mat_col Ucolp(U, p);
                    real_mat_col Ucolq(U, q);

                    real_vec xp(N), xq(N);
            
                    xp.assign(upp*Mcolp + uqp*Mcolq);
                    xq.assign(upq*Mcolp + uqq*Mcolq);
                    Mcolp.assign(xp);
                    Mcolq.assign(xq);

                    xp.assign(upp*Mrowp + uqp*Mrowq);
                    xq.assign(upq*Mrowp + uqq*Mrowq);
                    Mrowp.assign(xp);
                    Mrowq.assign(xq);

                    xp.assign(upp*Ucolp + uqp*Ucolq);
                    xq.assign(upq*Ucolp + uqq*Ucolq);
                    Ucolp.assign(xp);
                    Ucolq.assign(xq);
                }
            }
        }

        return s/sdiag;
    }


    // Takagi factorization: U^T M U = diagonal

    template<class realT, class L, class A>
    realT takagi_factorize(boost::numeric::ublas::
                           matrix<std::complex<realT>, L, A>& M,
                           boost::numeric::ublas::
                           matrix<std::complex<realT>, L, A>& U,
                           unsigned int maxiter = 256,
                           realT prec = 1e-15)
    {
        using boost::numeric::ublas::matrix;
        using boost::numeric::ublas::matrix_row;
        using boost::numeric::ublas::matrix_column;
        using boost::numeric::ublas::matrix_range;
        using boost::numeric::ublas::identity_matrix;
        using boost::numeric::ublas::vector;
        using boost::numeric::ublas::vector_range;
        using boost::numeric::ublas::prod;

        typedef std::complex<realT> complT;

        typedef matrix<complT, L, A> compl_mat;
        typedef vector<complT, A> compl_vec;

        typedef typename compl_mat::size_type size_type;

        typedef matrix_column<compl_mat> compl_mat_col;
        typedef matrix_row<compl_mat> compl_mat_row;

        size_type N = M.size1();    
        BOOST_UBLAS_SAME(M.size1(), U.size1());
        BOOST_UBLAS_SAME(N, U.size2());
        BOOST_UBLAS_SAME(N, M.size2());

        realT s=0, sdiag=0;
        size_type iter;
        size_type p,q;
        realT prec2 = 2.*prec/(N*(N-1));

        for(iter = 0; iter < maxiter; iter++) {
            // sum of off-diagonal squared elements
            s = 0;
            sdiag = 0.;
            for(p = 0; p < N; p++) {
                sdiag += modsq(M(p,p));
                for(q = p+1; q < N; q++)
                    s += modsq(M(p,q));
            }
            if(s < prec*sdiag) break;


            for(p = 0; p < N; p++) {
                for(q = p+1; q < N; q++) {

                    // eigenvalues of the p,q submatrix

                    complT mpp = M(p,p);
                    complT mqq = M(q,q);
                    complT mpq = M(p,q);

                    if(modsq(mpq) < prec2*sdiag) continue;

                    complT hp = mpp/mpq;
                    complT hq = mqq/mpq;

                    complT y(hp.real() + hq.real(),
                             hp.imag() - hq.imag());
                    y /= std::abs(y);
                    complT yc = std::conj(y);

                    realT b = 
                        (hp.real()-hq.real())*y.real() +
                        (hp.imag()+hq.imag())*y.imag();
                    realT r = -b/2.+std::sqrt(b*b/4.+1.);
                    realT c = 1./std::sqrt(1.+r*r);


                    // eigenvectors
                    
                    complT upp = c;
                    complT uqp = c*r*y;
                    complT upq = -c*r*yc;
                    complT uqq = c;
                    

                    // matrix multiplications
                    
                    compl_mat_col Mcolp(M, p);
                    compl_mat_col Mcolq(M, q);
                    compl_mat_row Mrowp(M, p);
                    compl_mat_row Mrowq(M, q);
                    compl_mat_col Ucolp(U, p);
                    compl_mat_col Ucolq(U, q);

                    compl_vec xp(N), xq(N);
            
                    xp.assign(upp*Mcolp + uqp*Mcolq);
                    xq.assign(upq*Mcolp + uqq*Mcolq);
                    Mcolp.assign(xp);
                    Mcolq.assign(xq);

                    xp.assign(upp*Mrowp + uqp*Mrowq);
                    xq.assign(upq*Mrowp + uqq*Mrowq);
                    Mrowp.assign(xp);
                    Mrowq.assign(xq);

                    xp.assign(upp*Ucolp + uqp*Ucolq);
                    xq.assign(upq*Ucolp + uqq*Ucolq);
                    Ucolp.assign(xp);
                    Ucolq.assign(xq);
                }
            }
        }

        return s/sdiag;
    }

}}


#endif


